# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import GridUtils
from Game import GameState

# values in grid are 1, 2, or 3 digit numbers representing stacks. The rightmost digit
# is the top stack and determines the owning player.

MAX_PIECES = 18
VALUE_BASE = 10
MAX_STACK = 3

def is_grid_connected(grid):
    """ Returns true if all stacks in the grid form a connected group """
    filled_pred = lambda pos,val: val>0
    return 1==len(list(GridUtils.connected_groups_with_predicate(grid, filled_pred, ortho=False)))

class AbandeGameState(GameState):
    def owner_for_value(self, val):
        return val % VALUE_BASE
        
    def stack_size_for_value(self, val):
        return self.player_pieces_for_value(0, val)
        
    def player_pieces_for_value(self, pnum, val):
        num = 0
        while val>0:
            if pnum==0 or val%VALUE_BASE==pnum:
                num += 1
            val = val // VALUE_BASE
        return num
        
    def stack_contents_for_value(self, val):
        """ yields players owning the pieces in the stack, top to bottom """
        while val>0:
            yield val % VALUE_BASE
            val = val // VALUE_BASE
        
    def total_player_pieces(self, pnum):
        return sum(self.player_pieces_for_value(pnum, val) for val in self.itervalues())
        
    def player_score(self, pnum):
        score = 0
        for pos,val in self.enumerate_all():
            if self.owner_for_value(val)==pnum:
                # only score if a neighbor is controlled by opponent
                if any(nval>0 and self.owner_for_value(nval)!=pnum for npos,nval in self.enumerate_neighbors(pos)):
                    score += self.stack_size_for_value(val)
        return score
        
    def successors(self):
        pnum = self.current_player

        # allow drop if less than max pieces, must be adjacent to another piece (unless board is empty)
        empty = all(val==0 for val in self.itervalues())
        pieces = self.total_player_pieces(pnum)
        if pieces<MAX_PIECES:
            for pos in self.enumerate_coords_with_value(0):
                if empty or any(nval>0 for np,nval in self.enumerate_neighbors(pos)):
                    nextgrid = self.grid.copy()
                    nextgrid[pos] = pnum
                    yield ((pos,), self.next_state_with_grid(nextgrid))
                    
        # look for captures: must move on top of opponent stack, and preserve single group
        # first player can't capture first move by opponent
        if pnum==1 and pieces<=1: return
        for pos,val in self.enumerate_all():
            if self.owner_for_value(val)==pnum:
                ssize = self.stack_size_for_value(val)
                if ssize>=MAX_STACK: continue
                for np,nval in self.enumerate_neighbors(pos):
                    if nval>0 and self.owner_for_value(nval)!=pnum and self.stack_size_for_value(nval)+ssize<=MAX_STACK:
                        # combine stacks and check for single group
                        nextgrid = self.grid.copy()
                        nextgrid[pos] = 0
                        if is_grid_connected(nextgrid):
                            nextgrid[np] = nval * (10**ssize) + val
                            yield ((pos,np), self.next_state_with_grid(nextgrid))
                
    def animations_for_move(self, move, newstate):
        import PGPieceAnimation
        if len(move)<=1:
            return GameState.animations_for_move(self, move, newstate)
        moveanim = PGPieceAnimation.move_animation(move[0], move[1])
        return PGPieceAnimation.AnimationGroup([moveanim])
        